"""Module for defining models for landmark detection."""

from .cholesky_hourglass import CholeskyHourglass, StackedCholeskyHourglass
from .coord_conv import AddCoordChannels, CoordConvLayer
from .hourglass import Hourglass, StackedHourglass
from .spatial_configuration_net import (
    OriginalSpatialConfigurationNet,
    OriginalSpatialConfigurationNet3d,
    ProbSpatialConfigurationNet,
    SpatialConfigurationNet,
)

__all__ = [
    "Hourglass",
    "StackedHourglass",
    "CholeskyHourglass",
    "StackedCholeskyHourglass",
    "SpatialConfigurationNet",
    "OriginalSpatialConfigurationNet",
    "OriginalSpatialConfigurationNet3d",
    "ProbSpatialConfigurationNet",
    "CoordConvLayer",
    "AddCoordChannels",
]



"""Preprocessing utils functions for images"""

import glob
import os
from os.path import join

import cv2
import numpy as np
import torch
from tqdm import tqdm  # type: ignore


def normalize(img: np.ndarray) -> np.ndarray:
    """
    Normalize image to 0-1 range

    Args:
        img (np.ndarray): Image to normalize

    Returns:
        np.ndarray: Normalized image
    """
    if img.dtype == np.uint16:
        return cv2.normalize(img, None, 0, 65535, cv2.NORM_MINMAX)  # type: ignore
    if img.dtype == np.uint8:
        return cv2.normalize(img, None, 0, 255, cv2.NORM_MINMAX)  # type: ignore
    raise TypeError("Image type not supported")


def preprocess_all(folder: str, output_folder: str, fun_name: str) -> None:
    """
    Apply a preprocessing function to all images in a folder.

    Args:
        folder (str): Folder containing images to preprocess
        output_folder (str): Folder to store the preprocessed images
        fun_name (str): Name of the function to apply. (Only normalize supported for now)
    """
    fun = globals()[fun_name]
    for file in tqdm(glob.glob(folder + r"/**/*.png", recursive=True)):
        img = cv2.imread(join(folder, file), -1)
        adj_img = fun(img)
        target_path = join(folder, file)
        target_path = target_path.replace(folder, output_folder)
        if os.path.exists(os.path.dirname(target_path)) is False:
            os.makedirs(os.path.dirname(target_path))
        cv2.imwrite(target_path, adj_img)


def extract_roi(
    imgs: torch.Tensor,
    roi_middle: torch.Tensor,
    landmarks: torch.Tensor,
    size: int | tuple[int, ...] = 256,
    spatial_dims: int = 2,
    ensure_dim=True,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Extract a region of interest around the landmarks in the image.

    Args:
        img (torch.Tensor): Image to extract the region of interest from.
        roi_middle (torch.Tensor): Middle of the region of interest.
        landmark (torch.Tensor): Landmark to transform to the region of interest.
        size (int | tuple[int, int]): Size of the region of interest.
        spatial_dims (int): Number of spatial dimensions of the image.
        ensure_dim (bool): Ensure the output has the same number of dimensions as the input.

    Returns:
        torch.Tensor: Region of interest around the landmarks.
        torch.Tensor: Landmarks in the region of interest.
        torch.Tensor: Upper left corner of the region of interest.
    """
    if spatial_dims == 2:
        return extract_roi_2d(imgs, roi_middle, landmarks, size, ensure_dim)
    elif spatial_dims == 3:
        return extract_roi_3d(imgs, roi_middle, landmarks, size, ensure_dim)
    raise ValueError("spatial_dims must be 2 or 3.")


def extract_roi_2d(
    imgs: torch.Tensor,
    roi_middle: torch.Tensor,
    landmarks: torch.Tensor,
    size: int | tuple[int, ...],
    ensure_dim=True,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Extract a region of interest around the landmarks in the image.

    Args:
        img (torch.Tensor): Image to extract the region of interest from. (C, H, W)
        roi_middle (torch.Tensor): Middle of the region of interest. (2)
        landmarks (torch.Tensor): Landmarks to transform to the region of interest. (2)
        size (int): Size of the region of interest.
        ensure_dim (bool): Ensure the output has the same number of dimensions as the input.

    Returns:
        torch.Tensor: Region of interest around the landmarks.
        torch.Tensor: Landmarks in the region of interest.
        torch.Tensor: Upper left corner of the region of interest.
    """
    # perform checks on the input
    if imgs.dim() != 3:
        raise ValueError("Input image must have 3 dimensions.")
    if landmarks.shape[0] != 2:
        raise ValueError("Landmarks must have 2 dimensions.")
    if roi_middle.shape[0] != 2:
        raise ValueError("ROI middle must have 2 dimensions.")
    # check if tuple is passed
    if isinstance(size, int):
        size = (size, size)
    if imgs.shape[-2] < size[0] or imgs.shape[-1] < size[1]:
        raise ValueError("Size must be smaller than the image.")
    H, W = imgs.shape[-2:]
    roi_height_lower = (roi_middle[0] - size[0] // 2).int()
    roi_height_upper = (roi_middle[0] + size[0] // 2).int()
    roi_width_lower = (roi_middle[1] - size[1] // 2).int()
    roi_width_upper = (roi_middle[1] + size[1] // 2).int()
    if ensure_dim:
        if roi_height_lower < 0:
            roi_height_upper -= roi_height_lower  # type: ignore
            roi_height_lower = 0  # type: ignore
        if roi_height_upper > H:
            roi_height_lower -= roi_height_upper - H  # type: ignore
            roi_height_upper = H  # type: ignore
        if roi_width_lower < 0:
            roi_width_upper -= roi_width_lower  # type: ignore
            roi_width_lower = 0  # type: ignore
        if roi_width_upper > W:
            roi_width_lower -= roi_width_upper - W  # type: ignore
            roi_width_upper = W  # type: ignore
    else:
        torch.clamp(roi_height_lower, 0, H)
        torch.clamp(roi_height_upper, 0, H)
        torch.clamp(roi_width_lower, 0, W)
        torch.clamp(roi_width_upper, 0, W)
    roi_upper_left_point = torch.tensor([roi_height_lower, roi_width_lower])
    landmarks_patch = landmarks - torch.tensor([roi_height_upper - size[0], roi_width_lower])
    return (
        imgs[..., roi_height_lower:roi_height_upper, roi_width_lower:roi_width_upper],
        landmarks_patch,
        roi_upper_left_point,
    )


def extract_roi_3d(
    imgs: torch.Tensor,
    roi_middle: torch.Tensor,
    landmarks: torch.Tensor,
    size: int | tuple[int, ...],
    ensure_dim=True,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    if imgs.dim() != 5:
        raise ValueError("Input image must have 5 dimensions.")
    if landmarks.dim() != 3:
        raise ValueError("Landmarks must have 3 dimensions.")
    if landmarks.size(1) != 3:
        raise ValueError("Landmarks must have 3 columns.")
    if landmarks.size(0) != imgs.size(0):
        raise ValueError("Number of landmarks must match the batch size.")
    if roi_middle.dim() != 2:
        raise ValueError("ROI middle must have 2 dimensions.")
    if roi_middle.size(1) != 3:
        raise ValueError("ROI middle must have 3 columns.")
    if roi_middle.size(0) != imgs.size(0):
        raise ValueError("Number of ROI middle must match the batch size.")
    if isinstance(size, int):
        size = (size, size, size)
    if size[0] % 2 == 0 or size[1] % 2 == 0 or size[2] % 2 == 0:
        raise ValueError("Size must be an odd number.")
    roi_front_upper_left_point = roi_middle[:, 0] - torch.tensor(size).unsqueeze(0) // 2
    roi_height_lower = roi_middle[:, 0] - size[0] // 2
    roi_height_upper = roi_middle[:, 0] + size[0] // 2
    roi_width_lower = roi_middle[:, 1] - size[1] // 2
    roi_width_upper = roi_middle[:, 1] + size[1] // 2
    roi_depth_lower = roi_middle[:, 2] - size[2] // 2
    roi_depth_upper = roi_middle[:, 2] + size[2] // 2
    landmarks = landmarks - roi_front_upper_left_point
    return (
        imgs[
            :,
            :,
            roi_height_lower:roi_height_upper,
            roi_width_lower:roi_width_upper,
            roi_depth_lower:roi_depth_upper,
        ],
        landmarks,
        roi_front_upper_left_point,
    )



"""Hourglass network.
Proposed in: https://github.com/princeton-vl/pytorch_stacked_hourglass/tree/master
Partially inspired by: "Stacked Hourglass Network for Robust Facial Landmark Localisation"
            - Yang et al. (2017)
"""

from typing import Sequence

import torch
from monai.networks.blocks import Convolution, ResidualUnit
from torch import nn


class Hourglass(nn.Module):
    """
    Hourglass network is a network with symmetrical encoder and decoder paths. The encoder path
    downsamples the input image while the decoder path upsamples the image. Skip connections are
    added between the encoder and decoder paths to preserve spatial information. This network is
    used for pose estimation.
        Proposed in: "Stacked Hourglass Networks for Human Pose Estimation" - Newell et al. (2016)

    Args:
        spatial_dims (int): number of spatial dimensions of the input image.
        in_channels (int): number of input channels.
        out_channels (int): number of output channels.
        channels (Sequence[int], optional): number of output channels for each convolutional layer.
        conv_block (nn.Module, optional): convolutional block to use. Defaults to ResidualUnit.
        pooling (nn.Module, optional): pooling layer to use. Defaults to nn.MaxPool2d.
        up_sample_mode (str, optional): upsampling mode. Defaults to 'nearest'.
    """

    # TODO: implment the out_channels

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        channels: Sequence[int] = [64, 128, 256, 512],
        conv_block: nn.Module = ResidualUnit,  # type: ignore[assignment]
        pooling: nn.Module = nn.MaxPool2d,  # type: ignore[assignment]
        up_sample_mode: str = "nearest",
    ):
        super().__init__()
        self.spatial_dims = spatial_dims
        self.in_channels = in_channels
        self.depth = len(channels)
        self.down_blocks = nn.ModuleList(
            [
                nn.Sequential(
                    conv_block(
                        spatial_dims=self.spatial_dims,
                        in_channels=in_channels,
                        out_channels=channels[0],
                    ),
                    pooling(2),
                )
            ]
        )
        self.up_blocks = nn.ModuleList(
            [
                nn.Sequential(
                    conv_block(
                        spatial_dims=self.spatial_dims,
                        in_channels=channels[-1],
                        out_channels=channels[-1],
                    ),
                    nn.Upsample(scale_factor=2, mode=up_sample_mode),
                )
            ]
        )
        self.skip_blocks = nn.ModuleList(
            [
                conv_block(
                    spatial_dims=self.spatial_dims,
                    in_channels=channels[0],
                    out_channels=channels[-1],
                )
            ]
        )

        for i, channel in enumerate(channels[1:]):
            self.down_blocks.append(
                nn.Sequential(
                    conv_block(
                        spatial_dims=self.spatial_dims,
                        in_channels=channels[i],
                        out_channels=channel,
                    ),
                    pooling(2),
                )
            )
            self.skip_blocks.append(
                conv_block(
                    spatial_dims=self.spatial_dims, in_channels=channel, out_channels=channels[-1]
                )
            )
            self.up_blocks.append(
                nn.Sequential(
                    conv_block(
                        spatial_dims=self.spatial_dims,
                        in_channels=channels[-1],
                        out_channels=channels[-1],
                    ),
                    nn.Upsample(scale_factor=2, mode=up_sample_mode),
                )
            )
        self.neck_block = nn.Sequential(
            conv_block(
                spatial_dims=self.spatial_dims, in_channels=channels[-1], out_channels=channels[-1]
            ),
            Convolution(
                spatial_dims=self.spatial_dims,
                in_channels=channels[-1],
                out_channels=channels[-1],
                kernel_size=1,
                adn_ordering="N",
                norm="batch",
            ),  # TODO: check if this is correct (if it makes sense)
            Convolution(
                spatial_dims=self.spatial_dims,
                in_channels=channels[-1],
                out_channels=channels[-1],
                kernel_size=1,
                adn_ordering="N",
                norm="batch",
            ),  # TODO: check if this is correct (if it makes sense)
        )
        self.last_layer = nn.Sequential(
            Convolution(
                spatial_dims=self.spatial_dims,
                in_channels=channels[-1],
                out_channels=channels[-1] // 2,
                kernel_size=1,
                adn_ordering="N",
                norm="batch",
            ),
            Convolution(
                spatial_dims=self.spatial_dims,
                in_channels=channels[-1] // 2,
                out_channels=out_channels,
                kernel_size=1,
                adn_ordering="",
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)
        Returns:
            output tensor of shape (batch_size, out_channels, *img_dims)
        """
        down_outputs = []
        for down_block, skip_block in zip(self.down_blocks, self.skip_blocks):
            x = down_block(x)
            down_outputs.append(skip_block(x))
        x = self.neck_block(x)
        for up_block, skip_output in zip(self.up_blocks[::-1], down_outputs[::-1]):
            x = up_block(x + skip_output)
        x = self.last_layer(x)
        return x


class StackedHourglass(nn.Module):
    """
    Stacked hourglass.

    Args:
        nb_stacks (int): number of hourglass modules to stack.
        spatial_dims (int): number of spatial dimensions of the input image.
        in_channels (int): number of input channels.
        out_channels (int): number of output channels.
        channels (Sequence[int], optional): number of output channels for each convolutional layer.

        up_sample_mode (str, optional): upsampling mode. Defaults to 'nearest'.
    """

    def __init__(
        self,
        nb_stacks: int,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        channels: Sequence[int] = [64, 128, 256, 512],
        conv_block: nn.Module = ResidualUnit,  # type: ignore[assignment]
        pooling: nn.Module = nn.MaxPool2d,  # type: ignore[assignment]
        up_sample_mode: str = "nearest",
    ):
        super().__init__()
        self.nb_stacks = nb_stacks
        # TODO: in the orignal paper they add a pre sequential block
        self.stacks = nn.ModuleList(
            [
                Hourglass(
                    spatial_dims=spatial_dims,
                    in_channels=in_channels,
                    out_channels=out_channels,
                    channels=channels,
                    conv_block=conv_block,
                    pooling=pooling,
                    up_sample_mode=up_sample_mode,
                )
                for _ in range(nb_stacks)
            ]
        )
        self.output_to_feature_block = nn.ModuleList(
            [
                Convolution(
                    spatial_dims=spatial_dims,
                    in_channels=out_channels,
                    out_channels=in_channels,
                    kernel_size=1,
                    adn_ordering="N",
                    norm="batch",
                )
                for _ in range(nb_stacks - 1)
            ]
        )

    def forward(self, x: torch.Tensor) -> list[torch.Tensor]:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)
        Returns:
            list of output tensors of shape (batch_size, out_channels, *img_dims)
        """
        outputs = []
        next_x = x
        for i, stack in enumerate(self.stacks):
            pred = stack(next_x)
            outputs.append(pred)
            if i < self.nb_stacks - 1:
                next_x = x + self.output_to_feature_block[i](pred)
        return outputs



"""
Utility functions for visualizing data.
"""

from typing import Optional, Sequence

import matplotlib.pyplot as plt
import numpy as np
from torch import nn

from landmarker.data import HeatmapDataset, LandmarkDataset, MaskDataset
from landmarker.heatmap.decoder import heatmap_to_coord, heatmap_to_multiple_coord
from landmarker.heatmap.generator import HeatmapGenerator
from landmarker.utils.utils import pixel_to_unit


def inspection_plot(
    ds: LandmarkDataset,
    idx: int | Sequence[int],
    heatmap_generator: Optional[HeatmapGenerator] = None,
    save_path: Optional[str] = None,
    fig_title: str = "Landmark Dataset Inspection Plot",
):
    """
    Plots the transformed image, heatmap, and original image with landmarks for the given dataset
    indices.

    Args:
        ds (LandmarkDataset): Dataset to inspect.
        idx (int | Iterable[int]): Indices of the dataset to inspect.
        heatmap_generator (HeatmapGenerator, optional): Heatmap generator to use. Defaults to None.
        save_path (str, optional): Path to save the plot to. If None, the plot is not saved.
            Defaults to None.
        fig_title (str, optional): Title of the figure. Defaults to "Landmark Dataset Inspection
            Plot".
    """
    if isinstance(idx, int):
        idx = [idx]

    fig = plt.figure(figsize=(15, 5 * len(idx)))
    # fig.suptitle(fig_title)

    subfigs = fig.subfigures(len(idx), 1, squeeze=False).flatten()
    for row, subfig in enumerate(subfigs):
        ds_idx = idx[row]
        batch = ds[ds_idx]
        landmark = batch["landmark"]
        landmarks_original = ds.landmarks_original[ds_idx]
        img_t = batch["image"]
        img = ds.image_loader(ds.img_paths[ds_idx])  # type: ignore
        if not isinstance(ds, (HeatmapDataset, MaskDataset)):
            batch = ds[ds_idx]
            if heatmap_generator:
                heatmap = heatmap_generator(landmark.unsqueeze(0)).squeeze(0)
                assert heatmap is not None, "Heatmap generator must return a heatmap."
            else:
                heatmap = None
        else:
            heatmap = batch["mask"]
        if img_t.shape[0] > 3:  # If more than 3 channels, remove other channels
            img_t = img_t[:3]
            img = img[:3]
        if img_t.shape[0] == 1:
            img_t = img_t[0]
            img = img[0]
            img = (img - img.min()) / (img.max() - img.min()) * 255
        else:
            img_t = img_t.permute(1, 2, 0)
            img = img.permute(1, 2, 0)
        img = img.detach().numpy().astype(np.uint8)
        if heatmap is not None:
            axs = subfig.subplots(nrows=1, ncols=3)
            axs[0].imshow(img_t)
            axs[1].imshow(img_t)
            axs[1].imshow(heatmap.detach().numpy().sum(axis=0), cmap="jet", alpha=0.5)
            if len(img.shape) == 2:
                axs[2].imshow(img, cmap="gray")
            else:
                axs[2].imshow(img)
            if len(landmark.shape) == 3:  # If multiple instances
                for i in range(landmark.shape[0]):
                    axs[0].scatter(landmark[i, :, 1], landmark[i, :, 0], c="r", s=5)
                for i in range(landmarks_original.shape[0]):
                    axs[2].scatter(
                        landmarks_original[i, :, 1], landmarks_original[i, :, 0], c="r", s=5
                    )
            else:
                axs[0].scatter(landmark[:, 1], landmark[:, 0], c="r", s=5)
                axs[2].scatter(landmarks_original[:, 1], landmarks_original[:, 0], c="r", s=5)
            axs[0].set_title("Transformed w/ landmarks")
            axs[1].set_title("Transformed w/ heatmap")
            axs[2].set_title("Original w/ landmarks")

        else:
            axs = subfig.subplots(nrows=1, ncols=2)
            axs[0].imshow(img_t)
            axs[1].imshow(img)
            if len(landmark.shape) == 3:
                for i in range(landmark.shape[0]):
                    axs[0].scatter(landmark[i, :, 1], landmark[i, :, 0], c="r", s=5)
                for i in range(landmarks_original.shape[0]):
                    axs[1].scatter(
                        landmarks_original[i, :, 1], landmarks_original[i, :, 0], c="r", s=5
                    )
            else:
                axs[0].scatter(landmark[:, 1], landmark[:, 0], c="r", s=5)
                axs[1].scatter(landmarks_original[:, 1], landmarks_original[:, 0], c="r", s=5)
            axs[0].set_title("Transformed w/ landmarks")
            axs[1].set_title("Original w/ landmarks")
        # subfig.suptitle(f"Image {ds_idx}")
    if save_path:
        plt.savefig(save_path)
    plt.show()


def prediction_inspect_plot(
    ds: LandmarkDataset,
    model: nn.Module,
    idx: int | Sequence[int],
    activation: nn.Module = nn.Identity(),
    save_path: Optional[str] = None,
    fig_title: str = "Landmark Prediction Model Inspection Plot",
):
    """
    Plots the transformed image, predicted heatmap, and original image with predicted and true
    landmarks for the given dataset indices.

    Args:
        ds (LandmarkDataset): Dataset to inspect.
        model (nn.Module): Model to use for prediction.
        idx (int | Sequence[int]): Indices of the dataset to inspect.
        activation (nn.Module, optional): Activation function to use. Defaults to nn.Identity().
        save_path (str, optional): Path to save the plot to. If None, the plot is not saved.
            Defaults to None.
        fig_title (str, optional): Title of the figure. Defaults to "Landmark Prediction Model
            Inspection Plot".
    """
    if isinstance(idx, int):
        idx = [idx]

    fig = plt.figure(figsize=(15, 5 * len(idx)))
    # fig.suptitle(fig_title)

    # Create len(idx)x1 subfigures
    subfigs = fig.subfigures(len(idx), 1, squeeze=False).flatten()
    for row, subfig in enumerate(subfigs):
        ds_idx = idx[row]
        batch = ds[ds_idx]
        img = ds.image_loader(ds.img_paths[ds_idx])  # type: ignore
        img_t = batch["image"]
        heatmap = activation(model(img_t.unsqueeze(0)))
        landmark_t = batch["landmark"].view((-1, batch["landmark"].shape[-1]))
        landmarks_original = ds.landmarks_original[ds_idx]
        landmarks_original = landmarks_original.view((-1, landmarks_original.shape[-1]))
        pred_landmarks_t = heatmap_to_coord(heatmap, method="local_soft_argmax")
        pred_landmarks = pixel_to_unit(
            pred_landmarks_t,
            pixel_spacing=None,
            dim=img_t.shape[-2:],
            dim_orig=batch["dim_original"],
            padding=batch["padding"],
        ).squeeze(0)
        pred_landmarks_t = pred_landmarks_t.squeeze(0)
        if img_t.shape[0] > 3:  # If more than 3 channels, remove other channels
            img_t = img_t[:3]
            img = img[:3]
        if img_t.shape[0] == 1:
            img_t = img_t[0]
            img = img[0]
            img = (img - img.min()) / (img.max() - img.min()) * 255
        else:
            img_t = img_t.permute(1, 2, 0)
            img = img.permute(1, 2, 0)
        img = img.float()
        img_t = img_t.float()
        heatmap = heatmap.squeeze(0)
        axs = subfig.subplots(nrows=1, ncols=3)
        axs[0].imshow(img_t)
        axs[0].scatter(landmark_t[:, 1], landmark_t[:, 0], c="b", s=5)
        axs[0].scatter(
            pred_landmarks_t.detach().numpy()[:, 1],
            pred_landmarks_t.detach().numpy()[:, 0],
            c="r",
            s=5,
        )
        axs[1].imshow(img_t)
        axs[1].imshow(heatmap.detach().numpy().sum(axis=0), cmap="jet", alpha=0.5)
        if len(img.shape) == 2:
            axs[2].imshow(img, cmap="gray")
        else:
            axs[2].imshow(img)
        axs[2].scatter(landmarks_original[:, 1], landmarks_original[:, 0], c="b", s=5)
        axs[2].scatter(
            pred_landmarks.detach().numpy()[:, 1], pred_landmarks.detach().numpy()[:, 0], c="r", s=5
        )

        axs[0].set_title("Transformed w/ landmarks")
        axs[1].set_title("Transformed w/ heatmap")
        axs[2].set_title("Original w/ landmarks")
        # subfig.suptitle(f"Image {ds_idx}")
    fig.legend(["True", "Predicted"])
    if save_path:
        plt.savefig(save_path)
    plt.show()


def prediction_inspect_plot_multi_instance(
    ds: LandmarkDataset,
    model: nn.Module,
    idx: int | Sequence[int],
    threshold: float = 0.5,
    window: int = 5,
    activation: nn.Module = nn.Identity(),
    save_path: Optional[str] = None,
    fig_title: str = "Landmark Prediction Model Inspection Plot",
):
    if isinstance(idx, int):
        idx = [idx]

    fig = plt.figure(figsize=(15, 5 * len(idx)))
    fig.suptitle(fig_title)

    # Create len(idx)x1 subfigures
    subfigs = fig.subfigures(len(idx), 1, squeeze=False).flatten()
    for row, subfig in enumerate(subfigs):
        ds_idx = idx[row]
        img = ds.image_loader(ds.img_paths[ds_idx])  # type: ignore
        batch = ds[ds_idx]
        img_t = batch["image"]
        landmark_t = batch["landmark"]
        landmarks_original = pixel_to_unit(
            landmark_t,
            pixel_spacing=None,
            dim=img_t.shape[-2:],
            dim_orig=batch["dim_original"],
            padding=batch["padding"],
        )
        dim_orig = batch["dim_original"]
        padding = batch["padding"]
        heatmap = activation(model(img_t.unsqueeze(0)))
        pred_landmarks_t_batch, _ = heatmap_to_multiple_coord(
            heatmap, method="argmax", threshold=threshold, window=window
        )
        pred_landmarks_t = pred_landmarks_t_batch[0]  # type: ignore
        true_landmarks_t = []
        true_landmarks_original = []
        pred_landmarks = []
        for c in range(landmark_t.shape[0]):
            true_landmarks_t_class = []
            true_landmarks_original_class = []
            pred_landmarks_class = []
            for i in range(landmark_t.shape[1]):
                if not landmark_t[c, i].isnan().any():
                    true_landmarks_t_class.append(landmark_t[c, i])
                    true_landmarks_original_class.append(landmarks_original[c, i])
            for i in range(len(pred_landmarks_t[c])):
                pred_landmarks_class.append(
                    pixel_to_unit(
                        pred_landmarks_t[c][i].unsqueeze(0),
                        pixel_spacing=None,
                        dim=img_t.shape[-2:],
                        dim_orig=dim_orig,
                        padding=padding,
                    )
                    .squeeze(0)
                    .squeeze(0)
                )
            true_landmarks_t.append(true_landmarks_t_class)
            true_landmarks_original.append(true_landmarks_original_class)
            pred_landmarks.append(pred_landmarks_class)
        if img_t.shape[0] > 3:  # If more than 3 channels, remove other channels
            img_t = img_t[:3]
            img = img[:3]
        if img_t.shape[0] == 1:
            img_t = img_t[0]
            img = img[0]
            img = (img - img.min()) / (img.max() - img.min()) * 255
        else:
            img_t = img_t.permute(1, 2, 0)
            img = img.permute(1, 2, 0)
        img = img.detach().numpy().astype(np.uint8)
        heatmap = heatmap.squeeze(0)
        axs = subfig.subplots(nrows=1, ncols=3)
        axs[0].imshow(img_t)
        axs[1].imshow(img_t)
        axs[1].imshow(heatmap.detach().numpy().sum(axis=0), cmap="jet", alpha=0.5)
        if len(img.shape) == 2:
            axs[2].imshow(img, cmap="gray")
        else:
            axs[2].imshow(img)
        for c in range(len(true_landmarks_t)):
            for i in range(max(len(true_landmarks_t[c]), len(pred_landmarks_t[c]))):
                if i < len(true_landmarks_t[c]):
                    axs[0].scatter(true_landmarks_t[c][i][1], true_landmarks_t[c][i][0], c="b", s=5)
                    axs[2].scatter(
                        true_landmarks_original[c][i][1],
                        true_landmarks_original[c][i][0],
                        c="b",
                        s=5,
                    )
                if i < len(pred_landmarks_t[c]):
                    axs[0].scatter(pred_landmarks_t[c][i][1], pred_landmarks_t[c][i][0], c="r", s=5)
                    axs[2].scatter(pred_landmarks[c][i][1], pred_landmarks[c][i][0], c="r", s=5)

        axs[0].set_title("Transformed w/ landmarks")
        axs[1].set_title("Transformed w/ heatmap")
        axs[2].set_title("Original w/ landmarks")
        # subfig.suptitle(f"Image {ds_idx}")
    fig.legend(["True", "Predicted"])
    if save_path:
        plt.savefig(save_path)
    plt.show()



"""
Datasets module.
"""

from .cepha400 import (
    get_cepha_heatmap_datasets,
    get_cepha_landmark_datasets,
    get_cepha_mask_datasets,
    get_cepha_patch_datasets,
)
from .endovis2015 import (
    get_endovis2015_heatmap_datasets,
    get_endovis2015_landmark_datasets,
)
from .plant_centroids import (
    get_plant_centroids_heatmap_datasets,
    get_plant_centroids_landmark_datasets,
    get_plant_centroids_mask_datasets,
)

__all__ = [
    "get_cepha_heatmap_datasets",
    "get_cepha_landmark_datasets",
    "get_cepha_mask_datasets",
    "get_cepha_patch_datasets",
    "get_endovis2015_heatmap_datasets",
    "get_endovis2015_landmark_datasets",
    "get_plant_centroids_heatmap_datasets",
    "get_plant_centroids_landmark_datasets",
    "get_plant_centroids_mask_datasets",
]



"""
Losses module for training landmark localization models.
"""

from .losses import (
    AdaptiveWingLoss,
    EuclideanDistanceVarianceReg,
    GaussianHeatmapL2Loss,
    GeneralizedNormalHeatmapLoss,
    MultivariateGaussianNLLLoss,
    NLLLoss,
    StackedLoss,
    StarLoss,
    WingLoss,
)

__all__ = [
    "GeneralizedNormalHeatmapLoss",
    "EuclideanDistanceVarianceReg",
    "MultivariateGaussianNLLLoss",
    "WingLoss",
    "AdaptiveWingLoss",
    "StarLoss",
    "StackedLoss",
    "GaussianHeatmapL2Loss",
    "NLLLoss",
]



"""Tests for the models module."""

import os
from functools import partial

import torch
from monai.networks.blocks import ResidualUnit

from landmarker.models.cholesky_hourglass import StackedCholeskyHourglass
from landmarker.models.coord_conv import CoordConvLayer
from landmarker.models.hourglass import Hourglass, StackedHourglass
from landmarker.models.spatial_configuration_net import (
    OriginalSpatialConfigurationNet,
    OriginalSpatialConfigurationNet3d,
    ProbSpatialConfigurationNet,
    SpatialConfigurationNet,
)
from landmarker.models.utils import LogSoftmaxND, SoftmaxND


def test_original_spatial_configuration_net():
    """Test OriginalSpatialConfigurationNet."""
    net = OriginalSpatialConfigurationNet(in_channels=1, out_channels=4)
    input_tensor = torch.randn((1, 1, 256, 256))
    expected_output_shape = (1, 4, 256, 256)

    output_tensor = net(input_tensor)

    assert output_tensor.shape == expected_output_shape

    net = OriginalSpatialConfigurationNet(in_channels=1, out_channels=4, init_weigths=True)

    output_tensor = net(input_tensor)

    assert output_tensor.shape == expected_output_shape


def test_original_spatial_configuration_net_3d():
    """Test OriginalSpatialConfigurationNet3d."""
    net = OriginalSpatialConfigurationNet3d(in_channels=1, out_channels=4)
    input_tensor = torch.randn((1, 1, 192, 96, 96))
    expected_output_shape = (1, 4, 192, 96, 96)

    output_tensor = net(input_tensor)

    assert output_tensor.shape == expected_output_shape


def test_spatial_configuration_net():
    """Test SpatialConfigurationNet."""
    net = SpatialConfigurationNet(
        spatial_dims=2,
        in_channels=1,
        out_channels=4,
        la_channels=(128, 128, 128, 128),
        la_strides=(2, 2, 2),
        la_num_res_units=2,
        la_norm="instance",
        la_dropout=0.0,
        sp_channels=128,
        sp_kernel_size=11,
        sp_downsample=16,
    )
    input_tensor = torch.randn((1, 1, 256, 256))
    expected_output_shape = (1, 4, 256, 256)

    output_tensor = net(input_tensor)

    assert output_tensor.shape == expected_output_shape

    net = SpatialConfigurationNet(
        spatial_dims=2,
        in_channels=1,
        out_channels=4,
        la_channels=(128, 128, 128, 128),
        la_strides=(2, 2, 2),
        la_num_res_units=2,
        la_norm="instance",
        la_dropout=0.0,
        sp_channels=128,
        sp_kernel_size=11,
        sp_downsample=16,
        sp_image_input=False,
    )

    output_tensor = net(input_tensor)

    assert output_tensor.shape == expected_output_shape


def test_prob_spatial_configuration_net():
    """Test ProbSpatialConfigurationNet."""
    net = ProbSpatialConfigurationNet(spatial_dims=2, in_channels=1, out_channels=4)
    input_tensor = torch.randn((1, 1, 256, 256))
    expected_output_shape = (1, 4, 256, 256)

    output_tensor = net(input_tensor)

    assert output_tensor.shape == expected_output_shape

    net = ProbSpatialConfigurationNet(
        spatial_dims=2, in_channels=1, out_channels=4, sp_image_input=True
    )

    output_tensor = net(input_tensor)

    assert output_tensor.shape == expected_output_shape

    net = ProbSpatialConfigurationNet(
        spatial_dims=2, in_channels=1, out_channels=4, sp_image_input=False
    )

    assert output_tensor.shape == expected_output_shape


def test_hourglass():
    """Test Hourglass.""" ""
    spatial_dims = 2
    in_channels = 1
    out_channels = 4
    channels = [16, 32, 64, 64]
    subunits = 3
    up_sample_mode = "nearest"

    model = Hourglass(
        spatial_dims=spatial_dims,
        in_channels=in_channels,
        out_channels=out_channels,
        channels=channels,
        conv_block=partial(ResidualUnit, norm="batch", subunits=subunits),
        up_sample_mode=up_sample_mode,
    )
    input_tensor = torch.randn((1, 1, 256, 256))
    expected_output_shape = (1, out_channels, 256, 256)

    output_tensor = model(input_tensor)

    assert output_tensor.shape == expected_output_shape


def test_stacked_hourglass():
    """Test StackedHourglass."""
    nb_stacks = 2
    spatial_dims = 2
    in_channels = 3
    out_channels = 4
    channels = [16, 32, 64, 64]
    subunits = 3
    up_sample_mode = "nearest"

    model = StackedHourglass(
        nb_stacks=nb_stacks,
        spatial_dims=spatial_dims,
        in_channels=in_channels,
        out_channels=out_channels,
        channels=channels,
        conv_block=partial(ResidualUnit, norm="batch", subunits=subunits),
        up_sample_mode=up_sample_mode,
    )
    input_tensor = torch.randn((1, 3, 256, 256))
    expected_output_shape = (1, 4, 256, 256)

    output_tensor = model(input_tensor)

    assert len(output_tensor) == 2
    for i in range(nb_stacks):
        assert output_tensor[i].shape == expected_output_shape


def test_stack_cholesky_hourglass():
    """
    Test StackedCholeskyHourglass.
    """
    nb_stacks = 2
    img_size = (256, 256)
    in_channels = 1
    out_channels = 4
    channels = [16, 32, 64, 64]
    subunits = 3
    up_sample_mode = "nearest"

    model = StackedCholeskyHourglass(
        nb_stacks=nb_stacks,
        img_size=img_size,
        in_channels=in_channels,
        out_channels=out_channels,
        channels=channels,
        conv_block=partial(ResidualUnit, norm="batch", subunits=subunits),
        up_sample_mode=up_sample_mode,
    )
    input_tensor = torch.randn((1, 1, 256, 256))
    expected_hm_output_shape = (1, 4, 256, 256)
    expected_cholesky_output_shape = (1, 4, 2, 2)

    hm_output_tensor, cholesky_output_tensor = model(input_tensor)

    assert len(hm_output_tensor) == nb_stacks
    assert len(cholesky_output_tensor) == nb_stacks
    assert hm_output_tensor[0].shape == expected_hm_output_shape
    assert cholesky_output_tensor[0].shape == expected_cholesky_output_shape


def test_coord_conv_layer_no_radial_channel():
    """
    Test CoordConvLayer without radial channel.
    """
    # create an instance of the CoordConvLayer class without radial channel
    layer = CoordConvLayer(
        spatial_dims=2,
        in_channels=3,
        out_channels=16,
        radial_channel=False,
        conv_block=partial(ResidualUnit, strides=4, kernel_size=7),
    )

    # create some dummy input tensors
    x = torch.randn(2, 3, 64, 64)

    # pass the input tensor through the layer
    out = layer(x)

    # check that the output has the correct shape
    assert out.shape == torch.Size([2, 16, 16, 16])


def test_coord_conv_layer_with_radial_channel():
    """
    Test CoordConvLayer with radial channel.
    """
    # create an instance of the CoordConvLayer class with radial channel
    layer = CoordConvLayer(
        spatial_dims=2,
        in_channels=3,
        out_channels=16,
        radial_channel=True,
        conv_block=partial(ResidualUnit, strides=1, kernel_size=3),
    )

    # create some dummy input tensors
    x = torch.randn(2, 3, 64, 64)

    # pass the input tensor through the layer
    out = layer(x)

    # check that the output has the correct shape
    assert out.shape == torch.Size([2, 16, 64, 64])


def test_coord_conv_layer_coord_channels_range():
    """
    Test that the values of the coordinate channels are within the range [-1, 1].
    """
    # create an instance of the CoordConvLayer class without radial channel
    layer = CoordConvLayer(
        spatial_dims=2,
        in_channels=3,
        out_channels=16,
        radial_channel=False,
        conv_block=partial(ResidualUnit, strides=1, kernel_size=3),
    )

    # create some dummy input tensors
    x = torch.randn(2, 3, 64, 64)

    # pass the input tensor through the layer
    out = layer.add_coord_channels(x)

    # check that the output values are within the range [-1, 1]
    assert out.shape == torch.Size([2, 5, 64, 64])
    assert (-1 <= out[:, 3:]).all() and (out[:, 3:] <= 1).all()


def test_softmax_nd():
    """Test the SoftmaxND class."""
    # Test for 2D case
    softmax_2d = SoftmaxND(spatial_dims=2)
    x = torch.randn(1, 3, 4, 4)
    output = softmax_2d(x)
    assert output.shape == x.shape
    assert torch.allclose(torch.sum(output, dim=(-2, -1)), torch.ones(1, 3))

    # Test for 3D case
    softmax_3d = SoftmaxND(spatial_dims=3)
    x = torch.randn(1, 3, 4, 4, 4)
    output = softmax_3d(x)
    assert output.shape == x.shape
    assert torch.allclose(torch.sum(output, dim=(-3, -2, -1)), torch.ones(1, 3))


def test_log_softmax_nd():
    """Test the LogSoftmaxND class."""
    # Test for 2D case
    log_softmax_2d = LogSoftmaxND(spatial_dims=2)
    x = torch.randn(1, 3, 4, 4)
    output = log_softmax_2d(x)
    assert output.shape == x.shape
    assert torch.allclose(torch.sum(torch.exp(output), dim=(-2, -1)), torch.ones(1, 3))

    # Test for 3D case
    log_softmax_3d = LogSoftmaxND(spatial_dims=3)
    x = torch.randn(1, 3, 4, 4, 4)
    output = log_softmax_3d(x)
    assert output.shape == x.shape
    assert torch.allclose(torch.sum(torch.exp(output), dim=(-3, -2, -1)), torch.ones(1, 3))



import torch
from torch import nn


class SoftmaxND(nn.Module):
    """
    Applies the Softmax function over N-dimensional input tensor.

    Args:
        spatial_dims (int): The number of spatial dimensions (2 or 3).

    Attributes:
        dim (tuple): The dimensions over which to apply the Softmax function.

    Methods:
        forward(x):
            Applies the Softmax function to the input tensor `x`.

    Example:
        >>> softmax_nd = SoftmaxND(spatial_dims=2)
        >>> input_tensor = torch.randn(1, 3, 4, 4)
        >>> output_tensor = softmax_nd(input_tensor)
    """

    def __init__(self, spatial_dims):
        super().__init__()
        self.dim = (-2, -1) if spatial_dims == 2 else (-3, -2, -1)

    def forward(self, x):
        max_val = x
        for d in self.dim:
            max_val, _ = torch.max(max_val, dim=d, keepdim=True)
        out = torch.exp(x - max_val)
        return out / torch.sum(out, dim=self.dim, keepdim=True)


class LogSoftmaxND(nn.Module):
    """
    Applies the LogSoftmax function over N-dimensional input.

    Args:
        spatial_dims (int): The number of spatial dimensions (2 or 3).

    Attributes:
        dim (tuple): The dimensions over which to apply the LogSoftmax function.
        spatial_dims (int): The number of spatial dimensions.

    Methods:
        forward(x):
            Applies the LogSoftmax function to the input tensor `x`.

            Args:
                x (torch.Tensor): The input tensor.

            Returns:
                torch.Tensor: The tensor after applying the LogSoftmax function.
    """

    def __init__(self, spatial_dims):
        super().__init__()
        self.dim = (-2, -1) if spatial_dims == 2 else (-3, -2, -1)
        self.spatial_dims = spatial_dims

    def forward(self, x):
        if self.spatial_dims == 2:
            out_max, _ = torch.max(x, dim=-1, keepdim=True)
            out_max, _ = torch.max(out_max, dim=-2, keepdim=True)
        else:
            out_max, _ = torch.max(x, dim=-1, keepdim=True)
            out_max, _ = torch.max(out_max, dim=-2, keepdim=True)
            out_max, _ = torch.max(out_max, dim=-3, keepdim=True)
        x_exp = torch.exp(x - out_max)
        x_exp_sum = torch.sum(x_exp, dim=self.dim, keepdim=True)
        return x - out_max - torch.log(x_exp_sum)
